## !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: __init__.py
# Project: parsers
# Author: Brian Cherinka
# Created: Sunday, 17th February 2019 12:40:21 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Sunday, 17th February 2019 3:44:50 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
from .base import Parser
from .sqla import SQLAParser




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: conftest.py
# Project: tests
# Author: Brian Cherinka
# Created: Thursday, 14th February 2019 6:04:55 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Sunday, 3rd March 2019 4:53:45 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

# import some fixtures we might need
from .models import model_a_factory, model_b_factory
from .database import session





# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: sqla.py
# Project: mixins
# Author: Brian Cherinka
# Created: Wednesday, 13th February 2019 3:49:07 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Sunday, 17th February 2019 3:45:21 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import inspect
import decimal
from sqlalchemy import func, bindparam
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.orm.util import AliasedClass
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import sqltypes, between
from operator import le, ge, gt, lt, eq, ne
from boolean_parser.parsers.base import BooleanParserException


opdict = {'<=': le, '>=': ge, '>': gt, '<': lt, '!=': ne, '==': eq, '=': eq}

class SQLAMixin(object):
    ''' A Mixin class to apply SQLAlchemy filter parsing

    This mixin adds a ``filter`` method to the parsed result which converts
    the parsed string object into an appropriate SQLAlchemy filter condition to be used
    in SQLAlchemy queries.

    '''

    def _check_models(self, classes):
        ''' Check the input modelclass format

        Checks if input classes is a module of modelclasses, a list of modelclasses
        or a single ModelClass and returns a list of ModelClass objects.

        Parameters:
            classes (object):
                A ModelClass module, list of models, or single ModelClass

        Returns:
            A list of ModelClasses
        '''

        # an entire module of classes
        if inspect.ismodule(classes):
            # an entire module of classes
            models = [i[1] for i in inspect.getmembers(
                classes, inspect.isclass) if hasattr(i[1], '__tablename__')]
        elif isinstance(classes, (list, tuple)):
            # a list of ModelClasses
            models = classes
        else:
            # assume a single ModelClass
            models = [classes]

        # check for proper modelclasses
        allmeta = all([isinstance(m, DeclarativeMeta) or isinstance(m, AliasedClass) for m in models])
        assert allmeta is True, 'All input classes must be of type SQLAlchemy ModelClasses'

        return models

    def _get_field(self, modelclass, field_name, base_name=None):
        ''' Return a SQLAlchemy attribute from a field name.

        Checks that a given model contains the named field.

        Parameters:
            modelclass (ModelClass):
                A SQLAlchemy ModelClass
            field_name (str):
                The database field name
            base_name (str):
                The database table name

        Returns:
            An SQLA instrumented attribute object
        '''

        field = None
        # Handle hierarchical field names such as 'parent.name'
        if base_name:
            # Match alias name
            if isinstance(modelclass, AliasedClass) and \
                    base_name == modelclass._aliased_insp.name.lower():
                field = getattr(modelclass, field_name, None)

            # Match table name
            if not field and base_name in modelclass.__tablename__:
                field = getattr(modelclass, field_name, None)
        else:
            # Handle flat field names such as 'name'
            field = getattr(modelclass, field_name, None)

        return field

    def filter(self, modelclass):
        ''' Return the condition as an SQLalchemy query filter condition

        Loops over all models and creates a filter condition for that model
        given the input filter parameters.

        Parameters:
            modelclass (objects):
                A set of ModelClasses to use in the filter condition

        Returns:
            A SQL query filter condition
        '''

        assert modelclass is not None, 'No input found'

        condition = None
        models = self._check_models(modelclass)

        for model in models:
            # get the SQLA instrumented attribute
            field = self._get_field(model, self.name, base_name=self.base)

            # if there is an attribute then break and use that model
            if field and hasattr(field, 'type') and hasattr(field, 'ilike'):
                break

        # raise if no attribute found
        if not field:
            raise BooleanParserException(f'Table {model.__tablename__} does not have field {self.name}')

        # produce the SQLA filter condition
        condition = self._filter_one(model, field=field, condition=condition)

        return condition

    def _filter_one(self, model, field=None, condition=None):
        ''' Create a single SQLAlchemy filter condition '''

        # if no field present return the original condition
        if not field:
            return condition

        # Prepare field and value
        lower_field, lower_value, lower_value_2 = self._bind_and_lower_value(field)

        # Handle postgresql arrays if any
        if isinstance(field.type, postgresql.ARRAY):
            condition = field.any(self.value, operator=opdict[self.operator])
            return condition

        # Handle scalar values

        # Return SQLAlchemy condition based on operator value
        # self.name is parameter name, lower_field is Table.parameterName
        elif self.operator == '<':
            condition = lower_field.__lt__(lower_value)

        elif self.operator == '<=':
            condition = lower_field.__le__(lower_value)

        elif self.operator == '>':
            condition = lower_field.__gt__(lower_value)

        elif self.operator == '>=':
            condition = lower_field.__ge__(lower_value)

        elif self.operator == '!=':
            field = getattr(model, self.name)
            value = self.value
            # Handle NULL values
            if value.lower() == 'null':
                condition = field.isnot(None)
            else:
                condition = lower_field.__ne__(lower_value)

        elif self.operator == '=' or self.operator == '==':
            # Handle string type
            if isinstance(field.type, sqltypes.TEXT) or \
                isinstance(field.type, sqltypes.VARCHAR) or \
                isinstance(field.type, sqltypes.String):
                field = getattr(model, self.name)
                value = self.value
                # Handle NULL values
                if value.lower() == 'null':
                    condition = field.is_(None)
                # if operator is straing equals, check accordingly
                elif self.operator == '==':
                    condition = lower_field.__eq__(lower_value)
                # otherwise, this operator maps to LIKE
                # x=5   ->  x LIKE '%5%' (x contains 5)
                # x=5*  ->  x LIKE '5%'  (x starts with 5)
                # x=*5  ->  x LIKE '%5'  (x ends with 5)
                elif value.find('*') >= 0:
                    value = value.replace('*', '%')
                    condition = lower_field.ilike(bindparam(self.fullname, value))
                else:
                    condition = lower_field.ilike('%' + bindparam(self.fullname, value) + '%')
            # For all other types, assume straight equality
            else:
                field = getattr(model, self.name)
                value = self.value
                # Handle NULL values
                if value.lower() == 'null':
                    condition = field.is_(None)
                else:
                    condition = lower_field.__eq__(lower_value)

        elif self.operator == 'between':
            # between condition
            condition = between(lower_field, lower_value, lower_value_2)

        elif self.operator in ['&', '|']:
            # bitwise operations
            condition = lower_field.op(self.operator)(lower_value) > 0

        return condition


    def _bind_and_lower_value(self, field):
        ''' Bind and lower the value based on field type'''

        lower_value_2 = None

        # get python field type
        ftypes = [float, int, decimal.Decimal]
        fieldtype = field.type.python_type

        # format the values
        value, lower_field = self._format_value(self.value, fieldtype, field)
        if hasattr(self, 'value2'):
            value2, lower_field = self._format_value(self.value2, fieldtype, field)

        # bind the parameter value to the parameter name
        boundvalue = bindparam(self.fullname, value, unique=True)
        lower_value = func.lower(boundvalue) if fieldtype not in ftypes else boundvalue
        if hasattr(self, 'value2'):
            boundvalue2 = bindparam(self.fullname, value2, unique=True)
            lower_value_2 = func.lower(boundvalue2) if fieldtype not in ftypes else boundvalue2

        return lower_field, lower_value, lower_value_2

    def _format_value(self, value, fieldtype, field):
        ''' Formats the value based on the fieldtype

        Formats the value to proper numreical type and lowercases
        the field for string fields.

        Parameters:
            value (str):
                The conditional value to format
            fieldtype (object):
                The python field type
            field (SQLA attribute):
                SQLA instrumented attribute

        Returns:
            The formatted value and lowercase field
        '''

        lower_field = field
        if fieldtype == float or fieldtype == decimal.Decimal:
            out_value = self._cast_value(value, datatype=float)
        elif fieldtype == int:
            out_value = self._cast_value(value, datatype=int)
        else:
            lower_field = func.lower(field)
            out_value = value

        return out_value, lower_field

    def _cast_value(self, value, datatype=float):
        ''' Cast a value to a specific Python type

        Parameters:
            value (int|float):
                A numeric value to cast to a float or integer
            datatype (object):
                The numeric cast function.  Can be either float or int.

        Returns:
            The value explicitly cast to an integer or float
        '''

        assert datatype in [float, int], 'datatype must be either float or int'
        try:
            if value.lower() == 'null':
                out = 'null'
            else:
                out = datatype(value)
        except (ValueError, SyntaxError):
            raise BooleanParserException(f'Field {self.name} expects a {datatype.__name__} value.  Received {value} instead.')
        else:
            return out




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: clause.py
# Project: actions
# Author: Brian Cherinka
# Created: Sunday, 17th February 2019 12:52:31 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Sunday, 17th February 2019 12:53:16 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

#
# Parsing Action classses
#


class BaseAction(object):
    ''' Base object representing a clause action

    An action to perform after parsing a string clause.  If set, actions run
    Actions are attached to clauses with the ``setParseAction`` on a given ``pyparsing``
    element.  :py:meth:`pyparsing.ParserElement.setParseAction` accepts a function or class to
    be applied to the ``pyparsing`` element during the parsing process.  Multiple actions can
    be attached by passing a list of functions or classes.  This class extracts the parsed data
    from the ``pyparsing`` element and makes it accessible as a variety of named attributes.

    Attributes:
        name: str
            The name of the extracted parameter
        base: str
            The base name of the extracted parameter, if any.
        fullname: str
            The full name of the extracted parameter as base + name
        data: dict
            The extracted parsed parameters from the pyparse clause
        parsed_clause: :py:class:`pyparsing.ParseResults`
            The original pyparsed results object
        input_clause: str
            The original input clause element
    '''

    def __init__(self, data):
        self.parsed_clause = data
        self.data = data[0].asDict()

        # parse the basic parameter name
        self._parse_parameter_name()

    def _parse_parameter_name(self):
        ''' parse the parameter name into a base + name '''
        name = self.data.get('parameter', None)
        assert name.count(
            '.') <= 1, f'parameter {name} cannot have more than one . '
        if '.' in name:
            self.base, self.name = name.split('.', 1)
        else:
            self.base = None
            self.name = name

    @property
    def fullname(self):
        ''' The full parameter name, including any base '''
        return f'{self.base}.{self.name}' if self.base else self.name


class Word(BaseAction):
    ''' Class action for handling word clauses

    This action performs a basic word parse.  The basic word
    is assigned as the ``name`` attribute.  Example word clauses:
    "alpha" or "alpha and beta or not charlie".

    '''

    def __init__(self, data):
        super(Word, self).__init__(data)

    def __repr__(self):
        return f'{self.name}'

    @property
    def input_clause(self):
        ''' Original input clause as a string '''
        return f'{self.fullname}'


class Condition(BaseAction):
    ''' Class action for handling conditional clauses

    This action performs a basic conditional parse.  The syntax for a
    conditional expressions is defined as "parameter operand value" or
    for "between" conditions, "parameter between value and value2".  The parameter name,
    operand, and parameter value is assigned as the ``name``, ``operator``, and
    ``value`` attribute, respectively.  Example conditional clauses:
    "x > 5" or "x > 5 and y < 3".  When using a "between" condition, e.g.
    "x between 3 and 5", an additional ``value2`` attribute is assigned the second
    parameter value.  For bitwise operands of '&' and '|', the value can also accept a negation
    prefix, e.g. "x & ~256", which evaluates to "x & -257".

    Allowed operands for conditionals are:
        '>', '>=, '<', '<=', '==', '=', '!=', '&', '|'

    In addition to the Base Attributes, the ``Condition`` action provides
    additional attributes containing the parsed condition parameters.

    Attributes:
        operator: str
            The operand used in the condition
        value: str
            The parameter value in the condition
        value2: str
            Optional second value, assigned when a "between" condition is used.

    '''

    def __init__(self, data):
        super(Condition, self).__init__(data)

        # extract the conditional operator and value
        self.operator = self.data.get('operator', None)
        self._extract_values()

    def __repr__(self):
        more = 'and' + self.value2 if hasattr(self, 'value2') else ''
        return self.name + self.operator + self.value + more

    @property
    def input_clause(self):
        ''' Original input clause as a string '''
        if self.operator == 'between':
            return f'{self.fullname} {self.operator} {self.value} and {self.value2}'
        else:
            return f'{self.fullname} {self.operator} {self.value}'

    def _extract_values(self):
        ''' Extract the value or values from the condition '''
        self.value = self.data.get('value', None)
        if not self.value:
            if self.operator == 'between':
                self.value = self._check_bitwise_value(self.data.get('value1'))
                self.value2 = self._check_bitwise_value(
                    self.data.get('value2'))

        self.value = self._check_bitwise_value(self.value)

    def _check_bitwise_value(self, value):
        ''' Check if value has a bitwise ~ in it

        Removes any bitwise ~ found in a value for a condition.
        If the operand is a bitwise & or |, convert the ~value to its
        integer appropriate.  E.g. ~64 -> -65.

        Parameters:
            value: str
                A string numerical value

        Returns:
            The str value or value converted to the proper bitwise negative
        '''

        if '~' in value:
            value = value.replace('~', '')
            if self.operator in ['&', '|']:
                value = str(-1 * (int(value)) - 1)

        return value




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: sqla.py
# Project: parsers
# Author: Brian Cherinka
# Created: Sunday, 17th February 2019 12:40:39 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Sunday, 17th February 2019 2:04:29 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
from boolean_parser.parsers import Parser
from boolean_parser.mixins import SQLAMixin
from boolean_parser.actions.clause import Condition
from boolean_parser.actions.boolean import BaseBool, BoolNot, BoolAnd, BoolOr
from boolean_parser.clauses import condition, between_cond
from sqlalchemy.sql import or_, and_, not_


sqlaop = {'and': and_, 'not': not_, 'or': or_}


class SQLACondition(SQLAMixin, Condition):
    ''' SQLAlchemy Conditional Action

    Subclasses from ``Condition`` and ``SQLAMixin`` to create an
    action that parses a string that represents a SQLAlchemy filter condition
    and allows for conversion from the parsed string result to a SQLAlchemy
    filter object.  For SQLAlchemy conditions, the syntax for a conditon expression
    is "database_table_name.parameter operand value" where "database_table_name.parameter"
    is a dotted syntax for `ModelClass.parameter` which maps to `dbtable_name.column_name`.
    For example, given a base ModelClass "TableModel" with parameter "x", that maps to a database
    table called "table" with column "x", the conditional expression "table.x < 4" parses into
    "{name: 'x', fullname: 'table.x', base: 'table', operator: '<', value: '4'}"

    '''
    pass


class SQLBoolBase(BaseBool):
    ''' Class for handling boolean logic joins for SQLALchemy filter expressions '''

    def filter(self, models):
        ''' Calls the filter method for each condition

        Parameters:
            models: list
                A list of SQLAlchemy ORM models
        '''
        conditions = [condition.filter(models)
                      for condition in self.conditions]
        return sqlaop[self.logicop](*conditions)


class SQLANot(BoolNot, SQLBoolBase):
    ''' SQLalchemy class for boolean Not '''


class SQLAAnd(BoolAnd, SQLBoolBase):
    ''' SQLalchemy class for boolean And '''


class SQLAOr(BoolOr, SQLBoolBase):
    ''' SQLalchemy class for boolean Or '''


class SQLAParser(Parser):
    ''' A SQLAlchemy boolean parser object

    A Parser class that provides a mechanism for converting a
    string conditional into a SQLAlchemy filter condition that can
    be passed into SQLAlchemy queries.  This parser contains a ``filter``
    method which accepts a list of SQLAlchemy Model classes used to identify
    and convert the parsed parameter name into a SQLAlchemy Instrumented Attribute.

    Example:
        >>> from boolean_parser.parsers import SQLAParser
        >>> from database.models import TableModel
        >>> from database import session
        >>>
        >>> # create the parser and parse a sql condition
        >>> res = SQLParser('table.x > 5 and table.y < 2').parse()
        >>> res
        >>> and_(x>5, y<2)
        >>>
        >>> # generate the sqlalchemy filter
        >>> ff = res.filter(TableModel)
        >>> print(ff.compile(compile_kwargs={'literal_binds': True}))
        >>> table.x > 5 AND table.y < 2
        >>>
        >>> # perform the sqlalchemy query
        >>> session.query(TableModel).filter(ff).all()
    '''
    _bools = [SQLANot, SQLAAnd, SQLAOr]


# Set new SQLAlchemy parse actions on conditions and build the Parser
clauses = [condition, between_cond]
actions = [SQLACondition, SQLACondition]
SQLAParser.build_parser(clauses=clauses, actions=actions)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: base.py
# Project: parsers
# Author: Brian Cherinka
# Created: Sunday, 17th February 2019 3:43:08 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Friday, 1st March 2019 2:18:11 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import copy
import six
import pyparsing as pp
from pyparsing import ParseException
from boolean_parser.actions.boolean import BoolNot, BoolAnd, BoolOr
from boolean_parser.clauses import condition, between_cond, words
from boolean_parser.actions.clause import Condition, Word


class BooleanParserException(Exception):
    pass


class Parser(object):
    ''' Core Parser class for parsing strings into objects

    A core Parser class that can parse strings into a set of objects
    based on a defined set of string clause elements, and actions to perform
    for each clause.
    '''
    _bools = [BoolNot, BoolAnd, BoolOr]
    _clauses = []
    _clause = None

    def __init__(self, value=None):
        self.original_input = value
        self._expression = None

        if self.original_input:
            self._expression = self.parse()

    @property
    def params(self):
        ''' The extracted parameters from the parsed string '''
        if isinstance(self._expression, Condition):
            return self._expression.fullname if self._expression else None
        return self._expression.params if self._expression else None

    @property
    def conditions(self):
        ''' The extracted conditions from the parsed string '''
        if isinstance(self._expression, Condition):
            return self._expression if self._expression else None
        return self._expression.conditions if self._expression else None

    def parse(self, value=None):
        ''' Parse a string conditional

        Calls ``parseString`` on the ``pyparsing`` clause element to parse the
        input string into a ``pyparsing.ParseResults`` object.

        Parameters:
            value: str
                The string expression to parse

        Returns:
            A pyparsing.ParseResults object

        Example:
            >>> from boolean_parser.parsers import Parser
            >>> pp = Parser()
            >>> pp.parse('x > 1')
            >>> x>1
        '''

        value = value or self.original_input
        assert value is not None, 'There must be some input to parse'
        assert isinstance(value, six.string_types), 'input must be a string'

        try:
            expression = self._parser.parseString(value)[0]
        except ParseException as e:
            raise BooleanParserException("Parsing syntax error ({0}) at line:{1}, "
                                         "col:{2}".format(e.markInputline(), e.lineno, e.col))
        else:
            self._expression = expression
            self.original_input = value
            return expression

    def __repr__(self):
        return f'<Parser(input="{self.original_input or ""}")>'

    @classmethod
    def build_parser(cls, clauses=None, actions=None, bools=None):
        ''' Builds a new boolean parser

        Constructs a new boolean Parser class given a set of clauses, actions,
        and boolean objects. Clauses are individual ``pyparsing`` elements that represent
        string clauses to pattern match on.  Actions are functions or classes set on each clause
        element that control how that clause is parsed.  See :ref:`clauses` for
        the available `pyparsing` clause elements.

        Assigns the default boolean classes, ``[BoolNot, BoolAnd, BoolOr]`` to the
        :py:func:`pyparsing.infixNotation` such that NOTs->ANDs->ORs.  If ``bools`` is
        specified instead, uses those object classes to handle boolean logic.  ``bools``
        must be a list of length 3 containing classes for boolean "not", "and", and "or" logic
        in that order.

        Parameters:
            clauses: list
                A list of pyparsing clause elements
            actions: list
                A list of actions to attach to each clause element
            bools: list
                A list of Boolean classes to use to handle boolean logic

        Example:
            >>> from boolean_parser.parsers import Parser
            >>> from boolean_parser.clauses import condition, words
            >>> from boolean_parser.actions.clause import Condition, Word
            >>>
            >>> # Assign the parsing order precedence for clauses
            >>> clauses = [condition, words]
            >>>
            >>> # Create a list of Actions for each clause in clauses
            >>> actions = [Condition, Word]
            >>>
            >>> # build the Parser with these clauses and actions
            >>> Parser.build_parser(clauses=clauses, actions=actions)
        '''

        # set clauses and actions
        clauses = clauses or cls._clauses
        assert clauses is not None, 'A list of clauses must be provided'
        cls.set_clauses(clauses)
        cls.build_clause()
        assert cls._clause is not None, 'A singular clause must be built from clauses'
        if actions:
            cls.set_parse_actions(clauses=clauses, actions=actions)

        # assign the combined clause to the recursive token pattern matcher
        where_exp = pp.Forward()
        where_exp <<= cls._clause

        # extract the Boolean precendent clauses
        bools = bools or cls._bools
        assert len(
            bools) == 3, 'there must be a set of "not, and, or" boolean precedent classes'
        bnot, band, bor = bools

        # build the expression parser
        cls._parser = pp.infixNotation(where_exp, [
            (pp.CaselessLiteral("not"), 1, pp.opAssoc.RIGHT, bnot),
            (pp.CaselessLiteral("and"), 2, pp.opAssoc.LEFT, band),
            (pp.CaselessLiteral("or"), 2, pp.opAssoc.LEFT, bor),
        ])

    @classmethod
    def set_parse_actions(cls, mapping=None, clauses=None, actions=None):
        ''' Attach actions to a pyparsing clause element

        ``pyparsing`` clause elements can have optional actions set with the
        ``setParseAction`` which control how each clause is parsed.  This maps a list
        of actions onto a list of clauses.  If ``mapping`` is used, it must be a list
        of tuples containing which action to map to which clause.  Otherwise
        ``clauses`` and ``actions`` must be provided as equal-length lists which contain,
        for each item, what action(s) to attach to the corresponding clause.

        Parameters:
            mapping: list of tuples
                A list of tuples containing a (clause, action) mapping.
            clauses: list
                A list of clauses
            actions: list
                A list of actions to attach to each clause

        Example:
            >>> from boolean_parser.parsers import Parser
            >>> from boolean_parser.clauses import condition, words
            >>> from boolean_parser.actions.clause import Condition, Word
            >>> clauses = [condition, words]
            >>> actions = [Condition, Word]
            >>> Parser.set_parse_actions(clauses=clauses, actions=actions)
        '''
        if not mapping:
            assert clauses and actions, 'clauses and actions must both be specified'
            assert isinstance(clauses, list), 'clauses must be a list'
            assert isinstance(actions, list), 'actions must be a list'
            assert len(clauses) == len(actions), 'clauses and actions must be the same length'
            mapping = zip(clauses, actions)
        else:
            assert isinstance(mapping, list), 'mapping must be a list'
            assert isinstance(mapping[0], (tuple, list)), 'mapping item must be a list or tuple'

        # requires clauses (a list)
        assert cls._clauses, 'class clauses must be set.  Call cls.set_clauses.'
        assert mapping is not None, 'a mapping between clauses and actions must be provided'

        # use reprs for list index check to bypass wonky list/equality clause comparisons
        clause_reprs = [repr(c) for c in cls._clauses]
        for item in mapping:
            clause, action = item
            assert repr(clause) in clause_reprs, 'clause must be included in list of class clauses'
            idx = clause_reprs.index(repr(clause))
            action = action if isinstance(action, (list, tuple)) else [action]
            cls._clauses[idx].setParseAction(*action)

    @classmethod
    def build_clause(cls, clauses=None):
        ''' Build a single clause from a list of clauses using pp.MatchFirst

        Merges a list of clauses into a single clause using :py:class:`pyparsing.MatchFirst`.
        This is equivalent to "clause = clause1 | clause2 | clause3`.  The clause precedence
        the Parser uses will be the order they appear in the list.  The default is to use
        the attached Parser._clauses list.

        Parameters:
            clauses: list
                A list of clauses to merge into a single clause
        '''
        clauses = clauses or cls._clauses
        assert isinstance(clauses, list), 'clauses must be a list'
        clauses = pp.MatchFirst(clauses)
        cls._clause = clauses

    @classmethod
    def set_clauses(cls, clauses):
        ''' Sets the list of clauses to use

        Parameters:
            clauses: list
                A list of clauses to attach to the Parser

        '''
        assert isinstance(clauses, list), 'clauses must be a list'
        cls._clauses = [copy.copy(c) for c in clauses]


# Set parse actions on conditions and build the Parser
clauses = [condition, between_cond, words]
actions = [Condition, Condition, Word]
Parser.build_parser(clauses=clauses, actions=actions)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: boolean.py
# Project: actions
# Author: Brian Cherinka
# Created: Sunday, 17th February 2019 12:52:38 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Sunday, 17th February 2019 12:52:53 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

#
# Boolean Precedent Actions
#


class BaseBool(object):
    ''' Base class for handling conditions joined by boolean logic operators

    This class handles the parsing of boolean logic within strings. The boolean
    classes are assigned to the :py:func:`pyparsing.infixNotation` in the order of boolean
    NOTS->ANDS->ORS, i.e. ``BoolNot``-> ``BoolAnd``-> ``BoolOr``.  For example,
    the string "x > 1 and y < 2", consisting of two conditions joined by a boolean "and" gets
    parsed into "and_(x>1, y<2)", represented as "BoolAnd(Condition1, Condition2)".
    During parsing, the class extracts all conditions and parameters joined by the boolean logic
    and makes them accessible as attributes.

    Attributes:
        params: list
            A list of extracted parameters from all conditions
        conditions: list
            A list of conditions contained within the boolean clause
        logicop: str
            The boolean logic operator used to join the conditions
    '''
    logicop = None

    def __init__(self, data):
        self._get_conditions(data[0])

    def _get_conditions(self, data):
        ''' Builds the list of conditions

        Iterators over the input parsed conditions and extracts
        pyparsed clauses from boolean logic joins. Extracts pyparsed
        conditions into the ``conditions`` parameter.

        Parameters:
            data: list
                A list of underlying conditions
        '''
        self.conditions = []
        for condition in data:
            # append to list if condition is not a Bool class
            if condition and condition != self.logicop:
                self.conditions.append(condition)

    @property
    def params(self):
        ''' The extracted parameters from a parsed condition '''
        params = []
        for condition in self.conditions:
            if isinstance(condition, BaseBool):
                params.extend(condition.params)
            else:
                params.append(condition.fullname)
        return list(set(params))

    def __repr__(self):
        strcond = ', '.join([repr(c) for c in self.conditions])
        return f'{self.logicop}_({strcond})'


class BoolNot(BaseBool):
    ''' Class for boolean Not logic '''
    logicop = 'not'


class BoolAnd(BaseBool):
    ''' Class for boolean And logic '''
    logicop = 'and'


class BoolOr(BaseBool):
    ''' Class for boolean Or logic '''
    logicop = 'or'



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: test_parse.py
# Project: tests
# Author: Brian Cherinka
# Created: Friday, 1st March 2019 3:11:43 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Friday, 1st March 2019 3:50:46 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import pytest
from boolean_parser import parse
from boolean_parser.parsers.base import BooleanParserException

#
# tests for the global basic parse function
#

# parametrization argument values
arg_values = [
    ('x > 5', 'x>5'),
    ('x > 5 and x < 10', 'and_(x>5, x<10)'),
    ('x between 5 and 10', 'xbetween5and10'),
    ('x > 5 or y < 3', 'or_(x>5, y<3)'),
    ('x > 5 or y < 3 and not z == 2', 'or_(x>5, and_(y<3, not_(z==2)))')
]


@pytest.mark.parametrize('value, exp',
                         [('stuff', 'stuff'),
                          ('stuff and things', 'and_(stuff, things)')] + arg_values)
def test_core_parse(value, exp):
    expr = parse(value, base='core')
    assert exp == repr(expr)


@pytest.mark.parametrize('value, exp', arg_values)
def test_sqla_parse(value, exp):
    expr = parse(value)
    assert exp == repr(expr)


@pytest.mark.parametrize('value', [('stuff'), ('stuff and things')])
def test_sqla_fails(value):
    with pytest.raises(BooleanParserException) as cm:
        parse(value)
    assert 'Parsing syntax error' in str(cm.value)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: clauses.py
# Project: boolean_parser
# Author: Brian Cherinka
# Created: Sunday, 17th February 2019 12:11:46 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Sunday, 17th February 2019 2:04:09 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import pyparsing as pp

# ------
# define base parser for words
word = pp.Word(pp.alphas).setResultsName('parameter')
words = pp.Group(word).setResultsName('words')

# ------
# define base parser for logical expressions
number = pp.Regex(r"[+-~]?\d+(:?\.\d*)?(:?[eE][+-]?\d+)?")
name = pp.Word(pp.alphas + '._', pp.alphanums +
               '._').setResultsName('parameter')
operator = pp.oneOf(['==', '<=', '<', '>', '>=', '=', '!=',
                     '&', '|']).setResultsName('operator')
value = (pp.Word(pp.alphanums + '-_.*') | pp.QuotedString('"')
         | number).setResultsName('value')
condition = pp.Group(name + operator + value).setResultsName('condition')

# ------
# define base parser for between expression
between_cond = pp.Group(name + pp.CaselessLiteral('between').setResultsName('operator') +
                        value.setResultsName('value1') + pp.CaselessLiteral('and') +
                        value.setResultsName('value2')).setResultsName('between_condition')

# -------
# define base parser for functions
ppc = pp.pyparsing_common

# parantheses
LPAR = pp.Suppress('(')
RPAR = pp.Suppress(')')

# function arguments
arglist = pp.delimitedList(number | (pp.Word(pp.alphanums + '-_') + pp.NotAny('=')))
args = pp.Group(arglist).setResultsName('args')
# function keyword arguments
key = ppc.identifier() + pp.Suppress('=')
values = (number | pp.Word(pp.alphas))
keyval = pp.dictOf(key, values)
kwarglist = pp.delimitedList(keyval)
kwargs = pp.Group(kwarglist).setResultsName('kwargs')
# build generic function
fxn_args = args + ',' + kwargs | pp.Optional(args, default='') + pp.Optional(kwargs, default='')
fxn_name = (pp.Word(pp.alphas)).setResultsName('name')
fxn = pp.Group(fxn_name + LPAR + fxn_args + RPAR).setResultsName('function')

# fxn condition
fxn_cond = pp.Group(fxn + operator + value).setResultsName('function_condition')

# fxn conditional expression
function_call = pp.Group(fxn_name + LPAR + condition + RPAR).setResultsName('function_call')
fxn_expr = pp.Group(function_call + operator + value).setResultsName('function_expression')

# cone fxn conditions
#cone_cond = copy.copy(fxn)
#cone_cond.setParseAction(ConeCondition)

# histogram fxn conditions
#hist_cond = copy.copy(fxn)
#hist_cond.setParseAction(HistCondition)


# create a list of usable constructed clauses
available_clauses = [words, condition, between_cond, fxn, fxn_cond, fxn_expr]



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: __init__.py
# Project: boolean_parser
# Author: Brian Cherinka
# Created: Wednesday, 13th February 2019 1:15:25 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Sunday, 17th February 2019 3:44:25 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

from boolean_parser.parsers import Parser
from boolean_parser.parsers import SQLAParser


__version__ = '0.1.3'


def parse(value, base='sqla'):
    ''' Convenience function to returned a parsed expression

    Returns a parsed string using one of the available Parsers
    in `boolean_parser`.  The ``base`` keyword argument can be used
    to select which ``Parser`` to use.  The availble bases are: "base",
    "sqla".  The default base is "sqla" which uses the :py:class:`boolean_parser.parsers.sqla.SQLAParser`.

    Parameters:
        value: str
            The string expression to evaluate and parse
        base: str:
            The base Parser to use.  Default is sqlalchemy parser.

    Returns:
        A parsed string using a :ref:`api-parsers` object.
    '''
    if base == 'base':
        return Parser(value).parse()
    elif base == 'sqla':
        return SQLAParser(value).parse()
    else:
        return Parser(value).parse()




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: test_clauses.py
# Project: tests
# Author: Brian Cherinka
# Created: Friday, 1st March 2019 3:08:52 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Sunday, 3rd March 2019 4:28:05 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import pytest
from boolean_parser.clauses import words, condition, between_cond, fxn, fxn_cond, fxn_expr

expdata = {'fxn1': {'name': 'test',
                    'args': ['1', '2', '3', 'hello'],
                    'kwargs': {'a': '5', 'stuff': 'there', 'force_check': 'True'}},
           'fxn2': {'name': 'test',
                    'args': ['1', '2', '3', 'hello'],
                    'kwargs': ''},
           'fxn3': {'name': 'test',
                    'args': '',
                    'kwargs': {'a': '5', 'stuff': 'there', 'force_check': 'True'}},
           'fxn4': {'name': 'test',
                    'args': '',
                    'kwargs': ''},
           'fxn_cond': {'function': {'name': 'test',
                                     'args': ['1', '2', '3'],
                                     'kwargs': {'a': '3', 'b': '4'}},
                        'operator': '>',
                        'value': '5'},
           'fxn_expr': {'function_call': {'name': 'test',
                                          'condition': {'parameter': 'x', 'operator': '>', 'value': '1'}},
                        'operator': '<=',
                        'value': '30'},
           'words': {'parameter': 'stuff'},
           'condition': {'parameter': 'a', 'operator': '>', 'value': '5'},
           'between_condition': {'parameter': 'a', 'operator': 'between', 'value1': '3', 'value2': '5'}
           }


@pytest.mark.parametrize('value, exp', [('stuff', expdata['words'])])
def test_words(value, exp):
    ''' test parsing of word clause '''
    res = words.parseString(value).asDict()
    assert 'words' in res
    assert res['words']['parameter'] == exp['parameter']


@pytest.mark.parametrize('value, exp', [('a > 5', expdata['condition'])])
def test_condition(value, exp):
    ''' test parsing of condition clause '''
    res = condition.parseString(value).asDict()
    assert 'condition' in res
    rc = res['condition']
    assert rc['value'] == exp['value']
    assert rc['parameter'] == exp['parameter']
    assert rc['operator'] == exp['operator']


def test_condition_has_default_action():
    ''' test that blank clauses have default actions '''
    assert condition.parseAction == []
    res = condition.parseString('x > 1').asDict()
    assert isinstance(res['condition'], dict)


@pytest.mark.parametrize('value, exp', [('a between 3 and 5', expdata['between_condition'])])
def test_between_condition(value, exp):
    res = between_cond.parseString(value).asDict()
    assert 'between_condition' in res
    bc = res['between_condition']
    assert bc['value1'] == exp['value1']
    assert bc['value2'] == exp['value2']
    assert bc['parameter'] == exp['parameter']
    assert bc['operator'] == exp['operator']


@pytest.mark.parametrize('value, exp',
                         [('test(1, 2, 3, hello, a=5, stuff=there, force_check=True)', expdata['fxn1']),
                          ('test(1, 2, 3, hello)', expdata['fxn2']),
                          ('test(a=5, stuff=there, force_check=True)', expdata['fxn3']),
                          ('test()', expdata['fxn4']),
                          ], ids=['all', 'args-only', 'kwargs-only', 'none'])
def test_function(value, exp):
    res = fxn.parseString(value).asDict()
    assert 'function' in res
    assert res['function']['name'] == exp['name']
    assert res['function']['args'] == exp['args']
    assert res['function']['kwargs'] == exp['kwargs']


@pytest.mark.parametrize('value, exp', [('test(1,2,3,a=3,b=4) > 5', expdata['fxn_cond'])])
def test_function_conditional(value, exp):
    res = fxn_cond.parseString(value).asDict()
    assert 'function_condition' in res
    fc = res['function_condition']
    assert fc['value'] == exp['value']
    assert fc['operator'] == exp['operator']
    assert fc['function'] == exp['function']

    assert fc['function']['name'] == exp['function']['name']
    assert fc['function']['args'] == exp['function']['args']
    assert fc['function']['kwargs'] == exp['function']['kwargs']


@pytest.mark.parametrize('value, exp', [('test(x > 1) <= 30', expdata['fxn_expr'])])
def test_function_expression(value, exp):
    res = fxn_expr.parseString(value).asDict()
    assert 'function_expression' in res
    fe = res['function_expression']
    assert fe['value'] == exp['value']
    assert fe['operator'] == exp['operator']
    assert fe['function_call'] == exp['function_call']

    assert fe['function_call']['name'] == exp['function_call']['name']
    assert fe['function_call']['condition'] == exp['function_call']['condition']

    fc = fe['function_call']['condition']
    ec = exp['function_call']['condition']
    assert fc['parameter'] == ec['parameter']
    assert fc['operator'] == ec['operator']
    assert fc['value'] == ec['value']



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: database.py
# Project: tests
# Author: Brian Cherinka
# Created: Sunday, 3rd March 2019 4:44:19 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Sunday, 3rd March 2019 4:49:52 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import pytest
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base


# global database setup
engine = create_engine('sqlite:///:memory:')
Session = scoped_session(sessionmaker(bind=engine))
Base = declarative_base(bind=engine)


@pytest.fixture(scope='function', autouse=True)
def session():
    session = Session()
    yield session
    session.rollback()
    session.close()




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: models.py
# Project: tests
# Author: Brian Cherinka
# Created: Friday, 15th February 2019 2:44:13 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Sunday, 3rd March 2019 4:47:18 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

from sqlalchemy import Column, Date, String, BigInteger, Integer, Float
from .database import Base, engine, Session
import factory
import factory.fuzzy
from pytest_factoryboy import register

import datetime

class ModelA(Base):
    __tablename__ = 'modela'
    pk = Column(BigInteger, primary_key=True)
    name = Column(String, nullable=False)
    nulls = Column(Integer, nullable=True)
    dates = Column(Date, nullable=False)
    x = Column(Integer, nullable=False)
    y = Column(Integer, nullable=False)

    def __repr__(self):
        return f'<ModelA(pk={self.pk},name={self.name},x={self.x},y={self.y})>'


class ModelB(Base):
    __tablename__ = 'modelb'
    pk = Column(BigInteger, primary_key=True)
    z = Column(Float, nullable=False)

    def __repr__(self):
        return f'<ModelB(pk={self.pk},z={self.z})>'


@register
class ModelAFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = ModelA
        sqlalchemy_session = Session   # the SQLAlchemy session object
    pk = factory.Sequence(lambda n: n)
    x = factory.Faker('pyint', min_value=0, max_value=20)
    y = factory.Faker('pyint', min_value=0, max_value=20)
    name = factory.fuzzy.FuzzyText(prefix='model', length=3)
    nulls = factory.Sequence(lambda n: None)
    dates = factory.Faker(
        'date_between_dates',
        date_start=datetime.date(2000, 1, 1),
        date_end=datetime.date(2020, 1, 1),
    )


@register
class ModelBFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = ModelB
        sqlalchemy_session = Session   # the SQLAlchemy session object
    pk = factory.Sequence(lambda n: n)
    z = factory.Faker('pyint', min_value=0, max_value=20)


Base.metadata.create_all(engine)
